# stolen from https://github.com/bybit-exchange/pybit/blob/master/pybit/exceptions.py
class FailedRequestError(Exception):
    """
    Exception raised for failed requests.

    Attributes:
        request -- The original request that caused the error.
        message -- Explanation of the error.
        status_code -- The code number returned.
        time -- The time of the error.
        resp_headers -- The response headers from API. None, if the request caused an error locally.
    """

    def __init__(self, request, message, status_code, time, resp_headers):
        self.request = request
        self.message = message
        self.status_code = status_code
        self.time = time
        self.resp_headers = resp_headers
        super().__init__(
            f"{message.capitalize()} (ErrCode: {status_code}) (ErrTime: {time})"
            f".\nRequest → {request}."
        )


from ._p2p_method import P2PMethod


class P2PMethods:
    GET_CURRENT_BALANCE = P2PMethod(
        "/v5/asset/transfer/query-account-coins-balance",
        "GET",
        [
            "accountType"
        ]
    )
    GET_ACCOUNT_INFORMATION = P2PMethod(
        "/v5/p2p/user/personal/info",
        "POST",
        []
    )
    GET_ADS_LIST = P2PMethod(
        "/v5/p2p/item/personal/list",
        "POST",
        []
    )
    GET_ONLINE_ADS = P2PMethod(
        "/v5/p2p/item/online",
        "POST",
        [
            "tokenId",
            "currencyId",
            "side"
        ]
    )




class P2PMethod:
    def __init__(
        self,
        url,
        http_method,
        required_params
    ):
        self.url = url
        self.http_method = http_method
        self.required_params = required_params



from typing import List


class Currency:
    def __init__(
        self,
        currencyId: str,
        exchangeId: str,
        id: str,
        orgId: str,
        scale: int
    ):
        self.currency_id = currencyId
        self.exchange_id = int(exchangeId)
        self.id = int(id)
        self.org_id = int(orgId)
        self.scale = scale


class Token:
    def __init__(
        self,
        exchangeId: str,
        id: str,
        orgId: str,
        scale: str,
        sequence: int,
        tokenId: str
    ):
        self.token_id = tokenId
        self.exchange_id = int(exchangeId)
        self.id = int(id)
        self.org_id = int(orgId)
        self.scale = scale
        self.sequence = sequence


class TradingPreferenceSet:
    def __init__(
        self,
        completeRateDay30: str,
        hasCompleteRateDay30: int,
        hasNationalLimit: int,
        hasOrderFinishNumberDay30: int,
        hasRegisterTime: int,
        hasUnPostAd: int,
        isEmail: int,
        isKyc: int,
        isMobile: int,
        nationalLimit: str,
        orderFinishNumberDay30: int,
        registerTimeThreshold: int
    ):
        self.complete_rate_day_30 = completeRateDay30
        self.has_complete_rate_day_30 = bool(hasCompleteRateDay30)
        self.has_national_limit = bool(hasNationalLimit)
        self.has_order_finish_number_day_30 = bool(hasOrderFinishNumberDay30)
        self.has_register_time = bool(hasRegisterTime)
        self.has_un_post_ad = bool(hasUnPostAd)
        self.is_email = bool(isEmail)
        self.is_kyc = bool(isKyc)
        self.id_mobile = bool(isMobile)
        self.national_limit = nationalLimit
        self.order_finish_number_day_30 = orderFinishNumberDay30
        self.register_time_threshold = registerTimeThreshold


class SymbolInfo:
    def __init__(
        self,
        buyAd,
        sellAd,
        buyFeeRate: str,
        currency: Currency,
        currencyId: str,
        currencyLowerMaxQuote: str,
        currencyMaxQuote: str,
        currencyMinQuote: str,
        exchangeId: str,
        id: str,
        itemDownRange: str,
        itemSideLimit: int,
        itemUpRange: str,
        kycCurrencyLimit: str,
        lowerLimitAlarm: int,
        orderAutoCancelMinute: int,
        orderFinishMinute: int,
        orgId: str,
        sellFeeRate: str,
        status: int,
        token: Token,
        tokenId: str,
        tokenMaxQuote: str,
        tokenMinQuote: str,
        tradeSide: int,
        upperLimitAlarm: int
    ):
        self.buy_ad = buyAd
        self.buy_fee_rate = buyFeeRate
        self.currency = currency
        self.currency_id = currencyId
        self.currency_lower_max_quote = float(currencyLowerMaxQuote)
        self.currency_max_quote = float(currencyMaxQuote)
        self.currency_lower_max_quote = float(currencyLowerMaxQuote)
        self.currency_min_quote = float(currencyMinQuote)
        self.exchange_id = int(exchangeId)
        self.id = int(id)
        self.item_down_range = float(itemDownRange)
        self.item_side_limit = itemSideLimit
        self.item_up_range = float(itemUpRange)
        self.kyc_currency_limit = float(kycCurrencyLimit)
        self.lower_limit_alarm = lowerLimitAlarm
        self.order_auto_cancel_minute = orderAutoCancelMinute
        self.order_finish_minute = orderFinishMinute
        self.org_id = int(orgId)
        self.sell_ad = sellAd
        self.sell_fee_rate = sellFeeRate
        self.status = status
        self.token = token
        self.token_id = tokenId
        self.token_max_quote = float(tokenMaxQuote)
        self.token_min_quote = float(tokenMinQuote)
        self.trade_side = tradeSide
        self.upper_limit_alarm = float(upperLimitAlarm)


class PaymentTemplateItem:
    def __init__(
        self,
        fieldName: str,
        labelDialect: str,
        placeholderDialect: str
    ):
        self.field_name = fieldName
        self.label_dialect = labelDialect
        self.placeholder_dialect = placeholderDialect


class PaymentConfig:
    def __init__(
        self,
        paymentDialect: str,
        paymentName: str,
        paymentTemplateItem: list,
        paymentType: int
    ):
        self.peyment_dialect = paymentDialect
        self.payment_name = paymentName
        self.paymentTemplateItem = [PaymentTemplateItem(**item) for item in paymentTemplateItem]
        self.payment_type = paymentType


class PaymentTerm:
    def __init__(
        self,
        accountNo: str,
        bankName: str,
        branchName: str,
        businessName: str,
        clabe: str,
        concept: str,
        debitCardNumber: str,
        firstName: str,
        id: str,
        lastName: str,
        mobile: str,
        payMessage: str,
        paymentConfig: dict,
        paymentExt1: str,
        paymentExt2: str,
        paymentExt3: str,
        paymentExt4: str,
        paymentExt5: str,
        paymentExt6: str,
        paymentTemplateVersion: float,
        paymentType: str,
        qrcode: str,
        realName: str,
        realNameVerified: bool,
        secondLastName: str,
        visible: int
    ):
        self.account_no = accountNo if accountNo.strip() else None
        self.bank_name = bankName if bankName.strip() else None
        self.branch_name = branchName if branchName.strip() else None
        self.business_name = businessName if businessName.strip() else None
        self.clabe = clabe if clabe.strip() else None
        self.concept = concept if concept.strip() else None
        self.debit_card_number = debitCardNumber if debitCardNumber.strip() else None
        self.first_name = firstName if firstName.strip() else None
        self.id = int(id)
        self.last_name = lastName if lastName.strip() else None
        self.mobile = mobile if mobile.strip() else None
        self.pay_message = payMessage if payMessage.strip() else None
        self.payment_config = PaymentConfig(**paymentConfig)
        self.payment_ext_1 = paymentExt1 if paymentExt1.strip() else None
        self.payment_ext_2 = paymentExt2 if paymentExt2.strip() else None
        self.payment_ext_3 = paymentExt3 if paymentExt3.strip() else None
        self.payment_ext_4 = paymentExt4 if paymentExt4.strip() else None
        self.payment_ext_5 = paymentExt5 if paymentExt5.strip() else None
        self.payment_ext_6 = paymentExt6 if paymentExt6.strip() else None
        self.payment_template_version = paymentTemplateVersion
        self.payment_type = paymentType
        self.qrcode = qrcode if qrcode.strip() else None
        self.real_name = realName if realName.strip() else None
        self.real_name_verified = realNameVerified
        self.second_last_name = secondLastName if secondLastName.strip() else None
        self.visible = bool(visible)


class MarketAd:
    def __init__(
        self,
        accountId: str,
        createDate: str,
        currencyId: str,
        executedQuantity: str,
        fee: str,
        finishNum: str,
        frozenQuantity: str,
        id: str,
        isOnline: bool,
        itemType: str,
        lastLogoutTime: str,
        lastQuantity: str,
        maxAmount: str,
        minAmount: str,
        nickName: str,
        orderNum: int,
        paymentPeriod: int,
        payments: List[str],
        premium: str,
        price: str,
        priceType: int,
        quantity: str,
        recentExecuteRate: int,
        recentOrderNum: int,
        remark: str,
        side: int,
        status: int,
        symbolInfo: SymbolInfo,
        tokenId: str,
        tokenName: str,
        tradingPreferenceSet: TradingPreferenceSet,
        userId: str,
        verificationOrderAmount: str,
        verificationOrderLabels: list,
        verificationOrderSwitch: bool,
        version: float,
        updateDate: str = None,
        subsidyAd: bool = None,
        paymentTerms: list = None,
        feeRate = None,
        authTag: List[str] = None,
        ban: bool = None,
        baned: bool = None,
        blocked: str = None,
        makerContact: bool = None,
        recommend: bool = None,
        recommendTag: str = None,
        userMaskId: str = None,
        userType: str = None,
        authStatus: int = None,
    ):
        self.account_id = int(accountId)
        self.auth_status = authStatus
        self.auth_tag = authTag
        self.ban = ban
        self.baned = baned
        self.blocked = blocked
        self.create_date = int(createDate)
        self.currency_id = currencyId
        self.eexecuted_quantity = float(executedQuantity) if executedQuantity.strip() else None
        self.fee = float(fee) if fee.strip() else None
        self.finish_num = finishNum
        self.frozen_quantity = float(frozenQuantity) if frozenQuantity.strip() else None
        self.id = int(id)
        self.is_online = isOnline
        self.item_type = itemType
        self.last_lagout_time = int(lastLogoutTime)
        self.last_quantity = float(lastQuantity) if lastQuantity.strip() else None
        self.maker_contract = makerContact
        self.max_amount = float(maxAmount)
        self.min_amount = float(minAmount)
        self.nickname = nickName
        self.order_num = orderNum
        self.payment_period = paymentPeriod
        self.payments = payments
        self.premium = bool(premium)
        self.price = float(price)
        self.quantity = float(quantity)
        self.recent_execute_rate = recentExecuteRate
        self.recommend = recommend
        self.remark = remark
        self.side = "buy" if int(side) == 0 else "sell"
        self.status = status
        self.symbol_info = symbolInfo
        self.token_id = tokenId
        self.trading_preference_set = tradingPreferenceSet
        self.user_id = userId
        self.user_mask_id = userMaskId
        self.user_type = userType
        self.verification_order_amount = int(verificationOrderAmount)
        self.verification_order_labels = verificationOrderLabels
        self.verification_order_switch = verificationOrderSwitch
        self.version = version
        self.price_type = priceType
        self.recent_order_num = recentOrderNum
        self.token_name = tokenName
        self.recommend_tag = recommendTag
        self.update_date = updateDate
        self.subsidy_ad = subsidyAd
        self.payment_terms = [PaymentTerm(**item) for item in paymentTerms] if paymentTerms else []
        self.fee_rate = feeRate


class CoinBalance:
    def __init__(
        self,
        bonus: str,
        coin: str,
        transferBalance: str,
        walletBalance: str
    ):
        self.bonus = float(bonus) if bonus.strip() else None
        self.name = coin
        self.transfer_balance = float(transferBalance) if transferBalance.strip else None
        self.wallet_balance = float(walletBalance) if walletBalance.strip() else None


class PrivilegeInfo:
    def __init__(self, name: str, data: str):
        self.name = name
        self.data = data


class AccountInfo:
    def __init__(
        self,
        accountCreateDays: int,
        accountId: str,
        averageReleaseTime: str,
        averageTransferTime: str,
        badAppraiseCount: int,
        blocked: str,
        canSubOnline: bool,
        contactConfig: bool,
        contactCount: int,
        curPrivilegeInfo: list,
        defaultNickName: bool,
        email: str,
        executeNum: int,
        firstTradeDays: int,
        goodAppraiseCount: int,
        goodAppraiseRate: str,
        hasUnPostAd: int,
        isOnline: bool,
        kycCountryCode: str,
        kycLevel: int,
        last30TradeCurrency: str,
        lastLogoutTime: str,
        lostRoleAffected: bool,
        mobile: str,
        nickName: str,
        openApiSwitch: int,
        orderNum: int,
        paymentCount: int,
        paymentRealNameUneditable: bool,
        realName: str,
        realNameEn: str,
        realNameMask: str,
        recentFinishCount: int,
        recentRate: float,
        recentTradeAmount: str,
        registerTime: str,
        totalFinishBuyCount: int,
        totalFinishCount: int,
        totalFinishSellCount: int,
        totalTradeAmount: str,
        userCancelCountLimit: int,
        userCurPrivilege: list,
        userId: str,
        userTag: list,
        userType: str,
        vipLevel: int,
        vipProfit: str,
        whiteFlag: int,
        authStatus: int = None,
    ):
        self.account_create_days = accountCreateDays
        self.account_id = int(accountId)
        self.auth_status = authStatus
        self.average_release_time = int(averageReleaseTime)
        self.average_transfer_time = int(averageTransferTime)
        self.bad_appraise_count = badAppraiseCount
        self.blocked = blocked
        self.can_sub_online = canSubOnline
        self.contact_config = contactConfig
        self.contact_count = contactCount
        self.cur_privilege_info = [PrivilegeInfo(**info) for info in curPrivilegeInfo]
        self.default_nickname = defaultNickName
        self.email = email if email.strip() else None
        self.execute_num = executeNum
        self.first_trade_days = firstTradeDays
        self.good_appraise_count = goodAppraiseCount
        self.good_appraise_rate = float(goodAppraiseRate)
        self.has_un_post_ad = bool(hasUnPostAd)
        self.is_online = isOnline
        self.kyc_country_code = kycCountryCode
        self.kyc_level = kycLevel
        self.last_30_days_trade_currency = last30TradeCurrency
        self.last_logout_time = int(lastLogoutTime)
        self.last_role_affected = lostRoleAffected
        self.mobile = mobile if mobile.strip() else None
        self.nickname = nickName
        self.open_api_switch = bool(openApiSwitch)
        self.order_num = orderNum
        self.payment_count = paymentCount
        self.payment_real_name_uneditable = paymentRealNameUneditable
        self.real_name = realName
        self.real_name_en = realNameEn
        self.real_name_mask = realNameMask
        self.recent_finish_count = recentFinishCount
        self.recent_rate = recentRate
        self.recent_trade_amount = float(recentTradeAmount) if recentTradeAmount.strip() else None
        self.register_time = int(registerTime)
        self.total_finish_buy_count = totalFinishBuyCount
        self.total_finish_count = totalFinishCount
        self.total_finish_sell_count = totalFinishSellCount
        self.total_trade_amount = float(totalTradeAmount) if totalTradeAmount.strip() else None
        self.user_cur_privilege = userCurPrivilege
        self.user_id = int(userId)
        self.user_tag = userTag
        self.user_type = userType
        self.vip_level = vipLevel
        self.vip_profit = vipProfit
        self.white_flag = whiteFlag
        self.user_cancel_count_limit = userCancelCountLimit



import asyncio
import base64
import hashlib
import hmac
import json
import logging
import os
import time
from datetime import datetime as dt
from datetime import timezone
from json import JSONDecodeError
from typing import Optional

import aiofiles
import aiohttp
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5

from ._exceptions import FailedRequestError
from ._p2p_method import P2PMethod

_SUBDOMAIN_TESTNET = "api-testnet"
_SUBDOMAIN_MAINNET = "api"
_DOMAIN_MAIN = "bybit"
_DOMAIN_ALT = "bytick"
_TLD_MAIN = "com"

class P2PManager:
    def __init__(
        self,
        api_key: str,
        api_secret: str,
        testnet: bool = False,
        recv_window: int = 5000,
        rsa: bool = False,
        alt_domain: bool = False,
        tld: str = "com"
    ):
        self._testnet = testnet
        self._api_key = api_key
        self._api_secret = api_secret
        self._recv_window = recv_window
        self._rsa = rsa
        self._alt_domain = alt_domain

        self._subdomain = _SUBDOMAIN_TESTNET if self._testnet else _SUBDOMAIN_MAINNET
        self._domain = _DOMAIN_ALT if self._alt_domain else _DOMAIN_MAIN
        self._tld = tld
        self._url = f"https://{self._subdomain}.{self._domain}.{self._tld}"

        self._session = None

    def _sign(
        self,
        use_rsa_authentication: bool,
        param_str: str,
        binary: bool = False
    ):
        def generate_hmac():
            hash = hmac.new(
                bytes(self._api_secret, "utf-8"),
                param_str.encode("utf-8"),
                hashlib.sha256
            )
            return hash.hexdigest()
        
        def generate_hmac_binary():
            hash = hmac.new(
                bytes(self._api_secret, "utf-8"),
                param_str,
                hashlib.sha256
            )
            return hash.hexdigest()
        
        def generate_rsa():
            hash = SHA256.new(param_str.encode("utf-8"))
            encoded_signature = base64.b64encode(
                PKCS1_v1_5.new(RSA.importKey(self._api_secret)).sign(
                    hash
                )
            )
            return encoded_signature.decode()
        
        def generate_rsa_binary():
            hash = SHA256.new(param_str)
            encoded_signature = base64.b64encode(
                PKCS1_v1_5.new(RSA.importKey(self._api_secret)).sign(
                    hash
                )
            )
            return encoded_signature.decode()
        
        if not use_rsa_authentication:
            if binary:
                return generate_hmac_binary()
            return generate_hmac()
        else:
            if binary:
                return generate_rsa_binary()
            return generate_rsa()
        
    def _generate_sign(self, payload, timestamp):
        sign_string = str(timestamp) + self._api_key + str(self._recv_window) + payload
        return self._sign(self._rsa, sign_string)
    
    def _generate_sign_binary(self, payload, timestamp):
        sign_string = f"{timestamp}{self._api_key}{self._recv_window}".encode() + payload
        return self._sign(self._rsa, sign_string, True)
    
    def _cast_values(
        self,
        params
    ):
        str_params = [
            "itemId",
            "side",
            "currency_id",
            # get_ad_detail
            "id",
            # get_ad_detail
            "id",
            "priceType",
            "premium",
            "price",
            "minAmount",
            "maxAmount",
            "remark",
            "actionType",
            "quantity",
            "paymentPeriod",
            # -> tradingPreferenceSet
            "hasUnPostAd",
            "isKyc",
            "isEmail",
            "isMobile",
            "hasRegisterTime",
            "registerTimeThreshold",
            "orderFinishNumberDay30",
            "completeRateDay30",
            "nationalLimit",
            "hasOrderFinishNumberDay30",
            "hasCompleteRateDay30",
            "hasNationalLimit",
            # get_orders
            "beginTime",
            "endTime",
            "tokenId",
            "currencyId",
            "sortType",
            "page",
            "coin",
            "amount",
            # get chat message
            "startMessageId"

            # switch active mode
            "workStatus",
            "memberId"
        ]
        int_params = [
            "positionIdx",
        ]

        self._cast_dict_recursively(params, str_params, int_params)

    def _cast_dict_recursively(self, dictionary, str_params, int_params):
        for key, value in dictionary.items():
            if value is None:
                continue
            if isinstance(value, dict):
                self._cast_dict_recursively(value, str_params, int_params)
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, dict):
                        self._cast_dict_recursively(item, str_params, int_params)
                    else:
                        if key in str_params and not isinstance(item, str):
                            value[i] = str(item)
                        elif key in int_params and not isinstance(item, int):
                            value[i] = int(item)
            elif isinstance(value, bool):
                dictionary[key] = value
            else:
                if key in str_params and not isinstance(value, str):
                    dictionary[key] = str(value)
                elif key in int_params and not isinstance(value, int):
                    dictionary[key] = int(value)

    def _generate_payload(
        self,
        http_method: str = "GET",
        params: dict = None
    ):
        http_method = http_method.upper()
        if http_method == "GET":
            payload = "&".join(
                [
                    str(k) + "=" + str(v)
                    for k, v in sorted(params.items())
                    if v is not None
                ]
            )
            return payload
        elif http_method == "POST":
            self._cast_values(params)
            return json.dumps(params)
        
    
    async def _get_session(self):
        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession()
        return self._session
    
    async def close_session(self) -> None:
        if self._session is None:
            return None

        await self._session.close()

    async def _request(
        self,
        method: P2PMethod,
        params: dict = {}
    ):
        try:
            missing_params = [p for p in method.required_params if p not in params]
            if missing_params:
                await self.close_session()
                raise ValueError(f"Missing required parameters: {', '.join(missing_params)}")
            
            for i in params.keys():
                if isinstance(params[i], float) and params[i] == int(params[i]):
                    params[i] = int(params[i])

            timestamp = int(time.time() * 10 ** 3)
            contentType = "application/json"

            if method.http_method == "FILE":
                filepath = params["upload_file"]
                boundary = "boundary-for-file"
                contentType = f"multipart/form-data; boundary={boundary}"
                filename = os.path.basename(str(filepath))
                mime_type = "image/png"
                
                async with aiofiles.open(filepath, 'rb') as f:
                    binary_data = await f.read()

                
                payload = (
                    f"--{boundary}\r\n"
                    f"Content-Disposition: form-data; name=\"upload_file\"; filename=\"{filename}\"\r\n"
                    f"Content-Type: {mime_type}\r\n\r\n"
                ).encode() + binary_data + f"\r\n--{boundary}--\r\n".encode()
                signature = self._generate_sign_binary(payload, timestamp)
            else:
                payload = self._generate_payload(
                    method.http_method,
                    params
                )
                signature = self._generate_sign(
                    payload,
                    timestamp
                )

            headers = {
                'X-BAPI-API-KEY': self._api_key,
                'X-BAPI-SIGN': signature,
                'X-BAPI-SIGN-TYPE': '2',
                'X-BAPI-TIMESTAMP': str(timestamp),
                'X-BAPI-RECV-WINDOW': str(self._recv_window),
                'Content-Type': contentType
            }

            endpoint = self._url + method.url

            session = await self._get_session()

            if method.http_method == "GET":
                response = await session.get(
                    endpoint + f"?{payload}" if payload != "" else "",
                    headers=headers,
                )
            elif method.http_method in ["POST", "FILE"]:
                response = await session.post(
                    endpoint,
                    headers=headers,
                    data=payload
                )
            else:
                return False, f"Unsupported HTTP method: {method.http_method}"
            
            if response.status != 200:
                if response.status == 403:
                    error_msg = "Access denied error. Possible causes: 1) your IP is located in the US or Mainland China, 2) IP banned due to ratelimit violation"
                elif response.status == 401:
                    error_msg = "Unauthorized. Possible causes: 1) incorrect API key and/or secret, 2) incorrect environment: Mainnet vs Testnet"
                else:
                    error_msg = f"HTTP status code is: {response.status}, expected: 200"

                await self.close_session()
                raise FailedRequestError(
                    request=f"{endpoint}: {payload}",
                    message=error_msg,
                    status_code=response.status,
                    time=dt.now(timezone.utc).strftime("%H:%M:%S"),
                    resp_headers=headers,
                )
            
            try:
                response_data = await response.json()
            except JSONDecodeError as ex:
                await self.close_session()
                raise FailedRequestError(
                    request=f"{endpoint}: {payload}",
                    message="Could not decode JSON.",
                    status_code=response.status,
                    time=dt.now(timezone.utc).strftime("%H:%M:%S"),
                    resp_headers=headers,
                )
            
            ret_code = "retCode"
            ret_msg = "retMsg"

            if ret_code not in response_data:
                ret_code = "ret_code"
            if ret_msg not in response_data:
                ret_msg = "ret_msg"

            if response_data[ret_code]:
                return False, (response_data[ret_code], response_data[ret_msg])

            else:
                return True, response_data["result"]
        except Exception:
            raise Exception



from typing import List, Tuple, Union

from ._classes import AccountInfo, CoinBalance, MarketAd
from ._p2p_helper import P2PMethods
from ._p2p_manager import P2PManager


class P2PRequests(P2PManager):

    async def get_current_balance(
        self,
        account_type: str = "FUND",
        with_bonus: bool = False,
        member_id: int | str = None,
        coins: List[str] = None
    ) -> Tuple[bool, Union[Tuple[int, str], Tuple[str, int, List[CoinBalance]]]]:
        """
        Obtain wallet balance, query asset information of each currency.
        By default, currency information with assets or liabilities of 0 is not returned.

        Args:
            account_type (str, optional): Account type. UNIFIED, FUND or CONTRACT.
            with_bonus (bool, optional): With bonus filter. Default False.
            member_id (int | str, options): User Id. It is required when you use master api key to check sub account coin balance.
            coins (List[str]): Coin names

        Returns:
            Tuple[str, int, List[CoinBalance]]: Success status and tuple. If success tuple is account_type, member_id and a list of CoinBalance. If not success tuple is retCode and retMsg.
        """

        params = {
            "accountType": account_type,
            "withBonus": "0" if not with_bonus else "1"
        }

        if member_id:
            params["memberId"] = member_id

        if coins:
            params["coin"] = ",".join(coins)

        status, data = await self._request(
            method=P2PMethods.GET_CURRENT_BALANCE,
            params=params
        )

        if not status:
            return status, data

        balances = [CoinBalance(**coin) for coin in data["balance"]]

        return status, (data["accountType"], int(data["memberId"]), balances)

    async def get_market_ads(
        self,
        amount: int | str = None,
        bulk_maker: bool = False,
        can_trade: bool = False,
        currency_id: str = "USD",
        item_region: int = 1,
        page: int | str = 1,
        payment: List[str] = [],
        payment_period: List[int] = [],
        side: str = "buy",
        size: int | str = 10,
        sort_type: str = "OVERALL_RANKING",
        token_id: str = "USDT",
        va_maker: bool = False,
        verification_filter: bool = False
    ) -> Tuple[bool, Union[Tuple[int, str], Tuple[int, List[MarketAd]]]]:
        """
        Get market ads from the P2P marketplace.

        Args:
            amount (int | str, optional): Searching amount.
            bulk_maker (bool, optional): Show only block advertisers.
            can_trade (bool, optional): Show only eligible ads.
            currency_id (str, optional): Currency id. For example: HKD, EUR, USD.
            item_region (int, optional): Item region.
            page (int | str, optional): Ads page.
            payment (List[str], optional): Payment method ids.
            payment_period (List[int], optional): Payment period filter.
            side (str, optional): 'buy' or 'sell'.
            sort_type (str, optional): Sorting filter. Options: OVERALL_RANKING, TRADE_VOLUME, TRADE_COMPLETION_RATE, TRADE_PRICE.
            token_id (str, optional): Token id, like USDT, BTC, ETH, or USDC.
            va_maker (bool, optional): Show only verified makers.
            verification_filter (bool, optional): Ads with no verification needed filter.

        Returns:
            Tuple[int, List[MarketAd]]: Success status and tuple. If success tuple is total_count and a list of MarketAd instances. If not success tuple is retCode and retMsg.
        """

        side = side.lower()

        params = {
            "tokenId": token_id,
            "currencyId": currency_id,
            "side": "0" if side == "buy" else "1",
            "verificationFilter": 2 if not verification_filter else 0,
            "vaMaker": va_maker,
            "page": str(page),
            "size": str(size),
            "itemRegion": item_region,
            "canTrade": can_trade,
            "bulkMaker": bulk_maker,
            "payment": payment,
            "paymentPeriod": payment_period,
            "sortType": sort_type,
        }

        if amount:
            params["amount"] = str(amount)

        status, data = await self._request(
            method=P2PMethods.GET_ONLINE_ADS,
            params=params
        )

        if not status:
            return status, data
        
        total_count = data["count"]
        items = data["items"]

        ads = [MarketAd(**item) for item in items]
        return status, (total_count, ads)

    async def get_account_information(
        self,
        **kwargs
    ) -> Tuple[bool, Union[Tuple[int, str], AccountInfo]]:
        """
        Get Account Information

        Return Tuple[bool, Union[Tuple[int, str], AccountInfo]]: Success status and AccountInfo instance if success. If not success tuple: retCode and retMsg.
        """

        status, data = await self._request(
            method=P2PMethods.GET_ACCOUNT_INFORMATION,
            params=kwargs
        )

        return status, AccountInfo(**data)

    async def get_ads_list(
        self,
        item_id: int | str = None,
        available: bool = False,
        side: str = "buy",
        token_id: str = "USDT",
        page: int | str = 1,
        size: int | str = 10,
        currency_id: str = "USD"
    ) -> Tuple[bool, Union[Tuple[int, str], Tuple[int, bool, List[MarketAd]]]]:
        """
        Get account ads list.

        Args:
            item_id (int | str, optional): Ad id.
            available (bool, optional): Available or not. Default False
            side (str, optional): Ad side, "buy" or "sell". Default "buy".
            size (int | str, optional): Page size. Default 10.
            currency_id (str, optional): Currency id, for example: HKD, USD, EUR. Defautl USD.
            token_id (str, optional): Token id, for example: USDT, USDC, BTC. Default USDT.
            page (int | str, optional): Page number, default 1.
            size (int | str, optional): Page size, default 10,
            currency_id (str, optional): Currency id, for example: HKD, USD, EUR. Default USD.
        
        Returns:
            Tuple[bool, Union[Tuple[int, str], Tuple[int, bool, List[MarketAd]]]]: Success status and tuple. If success tuple is ads count, hidden flag and list of MarketAd instances. If not succes tuple is retCode and retMsg.
        """

        params = {
            "side": "0" if side == "buy" else "1",
            "status": "2" if available else "1",
            "tokenId": token_id,
            "page": str(page),
            "size": str(size),
            "currency_id": currency_id
        }

        if item_id:
            params["itemId"] = str(item_id)

        status, data = await self._request(
            method=P2PMethods.GET_ADS_LIST,
            params=params
        )

        if not status:
            return status, data

        return status, (data["count"], data["hiddenFlag"], [MarketAd(**info) for info in data["items"]])
